#coding: utf-8;
############################################################################
#                                                                          #
#    This file 'args.py'                                                   #
#    is part of 'SPyRO: Simple Python Remote Objects'                      #
#    Copyright (C) 2004-2005 by Eric Sadit Tellez Avila                    #
#    Copyright (C) 2005-2006 by Eric Sadit Tellez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

__doc__ = """

This module is an attempt to increment the security of SPyRO, giving
tools to check the type of the arguments passed to functions.

Note: Remote arguments passed by reference are not checked properly, and must
be checked in the returned values.


The form to specify the argument's costraints are given by a format string, a type or class, or None

If the format is an string, you can specify an OR operation, the followoing values can be used:

i => int
f => float
c => complex
o => an instance
s => string
u => an unicode string
b => boolean data
n => None, the argument can be None
l => list
t => tuple
d => dictionary

If the format is a type, class or a tuple of clases, it will be used as
second argument of the 'isinstace' statement.

If the format is None, then it will not be checked

The format to check arguments must be given in two elements:
positional arguments and named arguments.


"""

import types

class Arguments:
    """
    Check if the arguments are valid against |fmt_args|.
    If any constraint is not valid, the function
    will return raise a TypeError and will set the appropiate string in the
    raised Exception.

    If the number of the given arguments are lesser that supposed
    it will raise TypeError too
    
    A proper check of arguments inside functions can be performed using
    the function 'local()' to retrieve local variables. And get the
    retrieve arguments
    
    |fmt_args| is a collection of tuples (argument_name, argument_value)

    Example
    ( ('arg1',fmt_arg1),
      ('arg2',fmt_arg2),
      ('arg3',fmt_arg3),
      ...
    )

    The arguments will be create two checking instances a list and a
    dictionary. The list checks positional arguments |args|, then when
    positional arguments are
    exhausted, named arguments |kwargs| will be checked with the dictionary
    validator.
    """
    def __init__(self, fmt_args):
        """ Compile an argument list to positional and keyword argument
            formats """
        self.named = named = {} 
        self.positional = pos = []
        for name, format in fmt_args:
            expformat = self._expand(format)
            pos.append(expformat)
            named[name] = expformat

    Types = {'i': types.IntType,
             'f': types.FloatType,
             'c': types.ComplexType,
             'o': types.InstanceType,
             's': types.StringType,
             'u': types.UnicodeType,
             'b': types.BooleanType,
             'n': types.NoneType,
             'l': types.ListType,
             'd': types.DictionaryType,
             }
    
    def _expand(self, fmt):
        """ Expands a format to an entity usable by isinstance or None"""
        exp = []
        if exp is None:
            return None
        elif isinstance(fmt, types.StringTypes):
            for x in fmt:
                exp.append(Arguments.Types[x])
            return tuple(exp)
        else:
            return fmt
    
    def _checkarg(self, fmt, arg, nameorpos):
        """ Check for an argument, sends a TypeError Exception
        if arguments doesn't accomplish constraint """
        if fmt is None:
            return
        #print fmt
        if not isinstance(arg, fmt):
            raise TypeError, "Argument '%s' has not a valid type '%s', valid types are: %s"%(nameorpos, type(arg), fmt)
        

    def check(self, args=(), kwargs={}):
        """ Check arguments """
        ipos = iter(self.positional)
        pos = 1
        for arg in args:
            try:
                self._checkarg(ipos.next(), arg, pos)
                pos = pos + 1
            except StopIteration:
                raise TypeError, "Argument '%s' does not exists in the format of arguments"%pos
        for name, arg in kwargs.items():
            try:
                self._checkarg(self.named[name], arg, name)
            except KeyError:
                raise TypeError, "Argument '%s' does not exists in the format of arguments"%name
        return True

def check(fmt, args, kwargs):
    argcheck = Arguments( fmt )
    argcheck.check(args, kwargs)

def testfun1(arg1, arg2, arg3):
    print type(arg1), type(arg2), type(arg3)

if __name__ == '__main__':

    fmt = (  ("arg1","if"),
             ("arg2","sn"),
             ("argx","snf") )

    argcheck = Arguments(  (  ("arg1","if"),
                              ("arg2","sn"),
                              ("argx","snf") ) )
    for args, kwargs in ( ( (1, None, 'sadit'), {'Fumofu':1} ),
                          ( (1, 1.1, 'sadit'), {}),
                          ( (1, 2, 'sadit'), {}),
                          ( (1, 'eric', 'sadit'), {})):
        try:                       
            print argcheck.check( args, kwargs )
            print "0-BUENO:", args
        except Exception, err:
            print "0-ERROR:", args, " -> ", err
            

        try:                       
            print check( fmt, args, kwargs )
            print "1-BUENO:", args
        except Exception, err:
            print "1-ERROR:", args, " -> ", err
