#coding: iso-8859-1;
############################################################################
#                                                                          #
#    This file 'formats.py'                                                  #
#    is part of 'SPyRO: Simple Python Remote Objects'                      #
#    Copyright (C) 2004-2005 by Eric Sadit Tellez Avila                    #
#    Copyright (C) 2005-2006 by Eric Sadit Tellez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

import warn
warn = warn.Warn("SPyRO.formats")
import pickle
import cPickle
import marshal
import sre
import SPyRO as spyro

import types
from types import DictType

def _set_booleans():
    BooleanType = types.IntType
    __builtins__['True'] = 1
    __builtins__['False'] = 0

try:
    BooleanType = types.BooleanType
    True
except:
    warn.warn("An old version of python or jython was detected")
    _set_booleans()
 
class SPyRO_Format:
    """ The spyro format minimum requeriments """
    # General methods
    def contenttype(self):
        """ Returns the content type to send menssages """
        return "application/xml"
    
    def _loads(self, msg, comm):
        """ Loads a msg. Its called for every load_* if the load_*
        method is not overloaded """
        raise "loads must be implemented"

    def _dumps(self, msg, comm):
        """ Dumps a msg. Its called for every dump_* if the dump_*
        method is not overloaded """
        raise "dumps must be implemented"

    # Client side
    def dump_control(self, msg, commclient):
        """ Creates a DEL|BYE request or any other request at low level """
        return self._dumps(msg, commclient)

    def dump_callMethod(self, msg, commclient):
        """ Creates the call method """
        return self._dumps(msg, commclient)

    def dump_setAttr(self, msg, commclient):
        """ Creates the set attribute message """
        return self._dumps(msg, commclient)

    def dump_getAttr(self, msg, commclient):
        """ Creates the get attribute message """
        return self._dumps(msg, commclient)

    def load_result(self, msg, commclient):
        return self._loads(msg, commclient)
    
    ########### Server side
    def load_request(self, msg, peerconn):
        return self._loads(msg, peerconn)
    
    def dump_result(self, msg, peerconn):
        """ Create a result message with |msg| object"""
        return self._dumps(msg, peerconn)

    def dump_error(self, msg, peerconn):
        """ Create a result message with |msg| object"""
        return self._dumps(msg, peerconn)

class PickleFormat(SPyRO_Format):
    """ This is the Interface to serialize (load and dumps) of the
    pickle format broker. It is quite simple, and it just uses
    the cPickle package.
    """
    def __init__(self, commlevel):
        """ Create a new pickle format manager """
        self.commlevel = commlevel

    def contenttype(self):
        return "application/octet-stream"

    def _loads(self, msg, comm):
        """ Load received message, returns python data structures """
        # warn.debug("***************** READ1:", msg)
        load = cPickle.loads(msg)
        # warn.debug(print "***************** READ2:", load)
        return load

    def _dumps(self, msg, comm):
        """ Dumps a message into data to send it across the network"""
        #warn.debug( "***************** WRITE1:", msg )
        dump = cPickle.dumps(msg,self.commlevel)
        #warn.debug( print "***************** WRITE2:", dump )
        return dump
	

class MarshalWrapper(SPyRO_Format):
    """ Calls a personalized marshaler with loads and dumps methods
    MarshalWrapper envolves the with _loads and _dumps methods
    """
    def contenttype(self):
        return "application/octet-stream"
    
    def __init__(self, marshal):
        self.marshal = marshal

    def _loads(self, msg, comm):
        return self.marshal.loads(msg)

    def _dumps(self, msg, comm):
        return self.marshal.dumps(msg)
    
# Registed formats

try:
    import xml.marshal.generic
except:
    warn.debug("Unable to use XML generic to communicate (with ",
               "xml.marshal.generic module)")
try:
    import xml.marshal.wddx
except:
    warn.debug("Unable to use XML WDDX to communicate (with ",
               "xml.marshal.wddx module)")

try:
    import SOAPpy
    _USE_SOAPpy = True
except:
    _USE_SOAPpy = False
    warn.debug("Unable to use SOAP to communicate (with SOAPpy module)")

class SOAP_SOAPpy(SPyRO_Format):
    """
    More info in /usr/lib/python2.4/site-packages/SOAPpy/SOAPBuilder.py
    """
    def __init__(self):
        pass

    def _dumps(self, msg, comm):
        #x = SOAPpy.buildSOAP(msg,method="SPyRO_execute")
        x = SOAPpy.buildSOAP(msg)
        #print x
        return x

    def _loads(self, msg, comm):
        x = SOAPpy.parseSOAP(msg)
        x = x.v1._asdict()
        if x.has_key("auth"):
            if x['auth']: x['auth'] = x['auth']._asdict()
            if x.has_key("kwargs"):
                try:
                    x['kwargs'] = x['kwargs']._asdict()
                except:
                    x['kwargs'] = {}
                x['args'] = x['args']._aslist()
        return x

try:
    import ZSI
    from cStringIO import StringIO
    _USE_ZSI = True
except:
    _USE_ZSI = False
    warn.debug("Unable to use SOAP to communicate (with ZSI module)")

class SOAP_ZSI(SPyRO_Format):
    def __init__(self):
        self.tcany = ZSI.TC.Any()

    def _dumps(self, msg, comm):
        o = StringIO()
        w = ZSI.SoapWriter(o)
        w.serialize(msg, self.tcany)
        #w.serialize(msg)
        w.close()
        s = o.getvalue()
        o.close()
        return s

    def _loads(self, msg, comm):
        ps = ZSI.parse.ParsedSoap(msg)
        return ps.Parse(self.tcany)

try:
    import xmlrpclib
except:
    warn.debug("Unable to use XMLRPC to communicate (with xmlrpclib module)")

class XMLRPCFormat(SPyRO_Format):
    """ XMLRPC Format,

    SPyRO Clients can use the transparent mode.
    Non-SPyRO clients must send request to any of the following functions

    ==> SPyRO_execute(options)
    Note: SPyRO_execute has an alias called 'execute' with the same arguments
    Execute a dictionary (xmlrpc struct) with all the information to perform
    a remote operation. The possible arguments depends of the type of the
    request, (See the SPyRO's Request class).
       The basic arguments or fields are:
        |objname| The object name (ID in the server)
        |attrname| The name of the requested attribute
        |reqtype| The type of the request. It can be 'BYE', 'GET', 'SET', 'DEL',
        CALL'
        |auth| The authentication object to the request

       Advanced arguments:
        |local_server| If Request to call a remote method and exists port we call
        the method with the pass the arguments by reference the server receives
        the name or id in the remote peer (the caller) and a port to connect.
        The IP address is obtained by the socket
        |auth| The authentication object
        |extra| Additional info to append to message, for example routed calls

       When calling methods:
       
        |args| Positional arguments, tuple
        |kwargs| Keyword arguments, dictionary

    
    Other specific functions are:
    ==> getAttr(object_name, attribute_name)
    Retrieves the value of the attribute |attribute_name| from the object
    |object_name|
    
    ==> setAttr(object_name, attribute_name, new_value)
    Set the value to the attribute |attribute_name| of the object |object_name|
    
    ==> callMethod(object_name, method_name, positional_args, keyword_args, options)
    Call remote methods
    |object_name| The name of the object
    |method_name| The name of the method
    |positional_args| The position arguments (an array or list)
    |keyword_args| The named arguments (a struct or dictionary)
    |options| Aditional arguments (see 'SPyRO_execute'
    """
    def __init__(self, encoding="UTF-8"):
        self.encoding = encoding

    def _loads(self, msg, comm):
        #warn.debug("LOAD:","**************" * 10, len(msg))
        #warn.debug(msg)
        args, methodname = xmlrpclib.loads(msg)
        #warn.debug("XXXXXXXX:",repr(args))
        return args[0]

    def _dumps(self, msg, comm):
        #warn.debug("DUMP:","==============" * 10)
        #warn.debug(msg)
        y = xmlrpclib.dumps((msg,) , methodname="SPyRO_execute", allow_none=1, encoding = self.encoding)
        return y

    ## server side dumps
    def load_request(self, msg, peerconn):
        x = {
            'auth': None, 'local_server': None,
            'rettype': spyro.SendByValue,
            'reqtype' : spyro.Request.ReqCallMethod,
            'args' : (), 'kwargs' : {}, 'extra': None
            }
        args, methodname = xmlrpclib.loads(msg)
        options = {}
        if methodname == 'SPyRO_execute' or methodname == 'execute':
            options = args[0]
        elif methodname == 'getAttr':
            x['objname'], x['attrname'], options = args
            x['reqtype'] = 'GET'
        elif methodname == 'setAttr':
            x['objname'], x['attrname'], x['value'], options = args
            x['reqtype'] = 'SET'
        elif methodname == 'callMethod':
            x['objname'], x['attrname'], x['args'], x['kwargs'], options = args
            x['reqtype'] = 'CALL'
        x.update(options)
        #warn.debug(x)
        return x
    
    def dump_result(self, msg, comm):
        return xmlrpclib.dumps((msg,), methodresponse=1,
                               allow_none=1, encoding = self.encoding)
    
    def dump_error(self, msg, comm):
        return self.dump_result(msg, comm)

    
class XMLRPCFormatLegacy(XMLRPCFormat):
    """ XMLRPC Format,
    Non - SPyRO clients must send request to procedure "SPyRO_execute" with
    the proper arguments (Request object properties)
    """
    def __init__(self, encoding="UTF-8"):
        self.encoding = encoding

    def _loads(self, msg, comm):
        #warn.debug("LOAD:","**************" * 10, len(msg))
        #warn.debug(msg)
        args, methodname = xmlrpclib.loads(msg)
        #warn.debug("XXXXXXXX:",repr(args))
        return args[0]

    def dump_control(self, msg, commclient):
        # fake, its not supported
        return xmlrpclib.dumps(('',''), methodname='SPyRO_control',
                               allow_none=1, encoding = self.encoding)

    def dump_callMethod(self, msg, commclient):
        xmlrpcargs = tuple([msg['objname']] + list(msg['args']))
        y = xmlrpclib.dumps(xmlrpcargs, methodname=msg['attrname'],
                            allow_none=1, encoding = self.encoding)
        return y
    
    ## server side dumps
    def load_request(self, msg, peerconn):
        args, methodname = xmlrpclib.loads(msg)
        x = {
            'auth': None, 'local_server': None,
            'rettype' : spyro.SendByValue,
            'reqtype' : spyro.Request.ReqCallMethod,
            'objname' : args[0],
            'args' : args[1:],
            'attrname' : methodname,
            'kwargs' : {}, 'extra': None,
            }
        # warn.debug("REQUEST:", x)
        return x
    
    def dump_result(self, msg, comm):
        return xmlrpclib.dumps((msg['retvalue'],), methodresponse=1,
                               allow_none=1, encoding = self.encoding)
    
    def dump_error(self, msg, comm):
        x = xmlrpclib.dumps(xmlrpclib.Fault(msg['errcode'],msg['retvalue']),
                            encoding = self.encoding)
        return x

    def load_result(self, msg, commclient):
        try:
            args, methodname = xmlrpclib.loads(msg)
            x = {'retvalue': args[0],
                 'rettype': spyro.SendByValue,
                 'errcode': spyro.Response.Successful}
        except xmlrpclib.Fault, err:
            x = {'retvalue': err.faultString,
                 'rettype': spyro.SendByValue,
                 'errcode': spyro.Response.ErrorExceptionThrowed}
        return x

def PrologChooseFormat(priorities, msg):
    """ Choose the best format between local priorities and
    the priorities in the msg
    if |priorities| is None the Prolog.Priorities is used
    """
    if not priorities: priorities = Prolog.Priorities
    allentries = Prolog.EntryReader.findall(msg)
    candidates = []
    for protocol, priority in allentries:
        try:
            candidates((priorities[protocol] * float(priority), protocol))
        except:
            # the protocol does not exists or the priority is not
            # a floating/int number
            pass
    candidates.sort()
    format = None
    try:
        format = candidates[-1][0]
    except:
        warn.debug("There is not available format to perform connection: ",
                   priorities, allentries)
    return format

class Prolog(SPyRO_Format):
    """ The prolog of the connection """
    EntryReader = sre.compile(r"(\w+):(\d.\d)\n?")
    Priorities = {
        'PICKLE2':1.0,
        'PICKLE1':0.9,
        'PICKLE0':0.8,
        'MARSHAL':0.7,
        'XMLRPC':0.6,
        'XMLRPC-Legacy':0.6,
        'XML':0.5,
        'WDDX':0.5,
        'XMLGENERIC':0.5,
        'SOAP':0.4,
        'SOAPpy':0.3,
        'SOAPZSI':0.2
        }

    def __init__(self, priorities):
        """ If |priorities| is None |Prolog.Priorities| is used instead"""
        if priorities is None:
            self.priorities = Prolog.Priorities
        else:
            self.priorities = priorities
        self.isserver = True
        
    def join(self,x):
        return "%s:%s"%x

    def _dumps(self, msg, comm):
        return "(Prolog)"

    def _loads(self, msg, comm):
        return newResponse(Response.Successful, msg, SendByValue)

    ## Server side
    def load_request(self, msg, peerconn):
        return Request(None, None, Request.ReqBye, None).__dict__
    
    def dump_result(self, msg, peerconn):
        return map("\n".join(map(self.join, self.priorities.items())))

    def dump_error(self, msg, peerconn):
        return self.dump_result(msg, peerconn)
    

class ORBFormats(DictType):
    """ The Object Request Broker Format class """

    def __init__(self):
        DictType.__init__(self)

    def init(self):
        self["MARSHAL"] = self.fmtMARSHAL
        self["PICKLE"] = self.fmtPICKLE
        self["PICKLE0"] = self.fmtPICKLE0
        self["PICKLE1"] = self.fmtPICKLE1
        self["PICKLE2"] = self.fmtPICKLE2
        self["XML"] = self.fmtXML
        self["WDDX"] = self.fmtWDDX
        self["XMLGENERIC"] = self.fmtXMLGENERIC
        self["XMLRPC"] = self.fmtXMLRPC
        self["XMLRPC-Legacy"] = self.fmtXMLRPCLegacy
        if _USE_SOAPpy:
            self["SOAP"] = self.fmtSOAPpy
        else:
            self["SOAP"] = self.fmtZSI
        self["SOAPpy"] = self.fmtSOAPpy
        self["SOAPZSI"] = self.fmtZSI
        self[""] = self.fmtDEFAULT
        self["GUESS"] = self.fmtGUESS

    def setFormat(self, name, generator):
        """ Register a new format, with name |name| and |interface| is function
        that returns a fully functional interface to manage the format (i.e. a
        new instance of the format manager) """        
        self.dict[name] = generator

    def fmtGUESS(self, peerconn):
        return Prolog(peerconn.server.priorities)
    
    def fmtXMLRPC(self, peerconn):
        return XMLRPCFormat()

    def fmtXMLRPCLegacy(self, peerconn):
        return XMLRPCFormatLegacy()

    def fmtSOAPpy(self, peerconn):
        return SOAP_SOAPpy()

    def fmtZSI(self, peerconn):
        return SOAP_ZSI()

    def fmtMARSHAL(self, peerconn):
        return MarshalWrapper(marshal)
    
    def fmtPICKLE(self, peerconn):
        return PickleFormat(0)

    def fmtPICKLE0(self, peerconn):
        return PickleFormat(0)

    def fmtPICKLE1(self, peerconn):
        return PickleFormat(1)

    def fmtPICKLE2(self, peerconn):
        return PickleFormat(2)

    def fmtXML(self, peerconn):
        return MarshalWrapper(xml.marshal.generic)

    def fmtWDDX(self, peerconn):
        return MarshalWrapper(xml.marshal.wddx)

    def fmtXMLGENERIC(self, peerconn):
        return self.fmtXML(peerconn)

    def fmtDEFAULT(self, peerconn):
        return self.fmtXMLRPC(peerconn)

    def getFormat(self, name, peerconn):
        """ Retrieve a format broker |name| """
        return self[name](peerconn)

