#coding: iso-8859-1;
############################################################################
#                                                                          #
#    This file 'threadutils/ThreadPool.py'                                 #
#    is part of 'SPyRO: Simple Python Remote Objects'                      #
#    Copyright (C) 2004-2005 by Eric Sadit Téllez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

import Queue
import thread, threading
from threading import Thread, Condition
import sys
import time
import traceback
import warn

warn = warn.Warn("threadutils.ThreadPool")

try:
    True
except:
    __builtins__['True'] = 1
    __builtins__['False'] = 0

class RunThread(Thread):
    def __getstate__(self):
        return {}

    def __del__(self):
        if not self.__closed:
            self.stop()
    
    def __init__(self, pool):
        Thread.__init__(self)
        self.cond=Condition()
        self.putfunction = pool.put
        self.exit = False
        self.start()
        self.site = None
        self.exception = None
        self.pool = pool
        self.__closed = False
        
    def prepare(self, putfunction, endnotify_function,  function, args, argd):
        self.function = function
        self.args = args
        self.argd = argd
        self.exception = None
        self.putfunction = putfunction
        self.endnotify = endnotify_function
        self.cond.acquire()
        self.cond.notify()
        self.cond.release()

    def stop(self):
        self.cond.acquire()
        self.exit = True
        self.cond.notify()
        self.cond.release()
        self.__closed = True
    
    def run(self):
        while True:
            self.cond.acquire()
            self.putfunction(self)
            if not self.exit: self.cond.wait()
            self.cond.release()
            if self.exit: break
            exc, res = None, None
            try:
                res=self.function(*self.args,**self.argd)
            except Exception, e:
                warn.warn("Exception in ThreadPool: %s\n"%(e),
                          self.args, self.argd)
                exc = e
            if self.endnotify:
                if callable(self.endnotify):
                    self.endnotify(res, exc)
                else:
                    self.endnotify[0](res, exc, self.endnotify)

class ThreadPoolMixIn:
    def _init_(self, maxthreads, can_grow):
        self.maxthreads = maxthreads
        self.cond = Condition()
        self.wait = False
        self.cnt = self.maxthreads
        self.can_grow = can_grow

    def put_block(self, obj):
        """ Returns the obj thread to the pool, intern method """
        self.cond.acquire()
        #print "AA>", self.qsize(), self.wait
        self.put(obj)
        #print "BB>", self.qsize(), self.wait
        if self.wait and self.cnt == 0:
            self.cond.notify()
            self.wait = False
        self.cond.release()

    def begin(self, cnt=None):
        """ Begin a synchronous block. Block's inside |cnt| threads are
        executed, but the program control will wait it
        in the block (in the self.end call). If |cnt| is not present,
        self.maxthreads is used"""
        self.cond.acquire()
        if cnt:
            self.cnt = cnt
        else:
            self.cnt = self.maxthreads
        self.wait = True
        self.cond.release()

    def end(self):
        """ Wait to finish all the threads in this pool, since the
        self.begin call """
        self.cond.acquire()
        if self.wait: self.cond.wait()
        self.cond.release()
    
    def execute(self, endfunction, function, *args, **argd):
        """ Execute |function| into a diferent thread with args (positional
        arguments) and argd (named arguments).
        If |endfunction| is diferent of None:
           If it is callable: is called with is called with two arguments. The
           first argument is the result, the second is the
           exceptionCatched (if any).
           In other words: endfunction(res, exc)
           
           If |endfunction| is not callable: the first item is called as a
           function and |endfunction| is passed as third argument. The first
           two arguments are equal to the previous case.
           In other words: endfunction[0](res, exc, endfunction)
        |result| is
        the result value of calling |function|, and |exceptionCatched| is not
        None if an exception was catched while |function| was running """
        if self.can_grow:
            try:
                runthread = self.get_nowait()
            except Queue.Empty:
                runthread = RunThread(self)
                runthread = self.get()
        else:
            #warn.stack("Trying to get a new runthread")
            runthread=self.get()
            #warn.stack("Successful")
        if self.wait:
            self.cond.acquire()
            self.cnt = self.cnt - 1
            self.cond.release()
            runthread.prepare(self.put_block,endfunction,function,args,argd)
        else:
            runthread.prepare(self.put, endfunction, function, args, argd)

    def close(self):
        pass

#class ThreadPool(Queue.Queue, ThreadPoolMixIn):
class ThreadPool(ThreadPoolMixIn, Queue.Queue):
    """ Create at most |maxthreads| to retrieve sites. It must be used with
    Queue |get| (get a ThreadFetch) and |put| (return the thread) methods. """
    def __getstate__(self):
        return {}
    
    def __init__(self, maxthreads, can_grow = False):
        """ Create a pool of threads """
        #Queue.Queue.__init__(self, maxthreads)
        if can_grow:
            Queue.Queue.__init__(self)
        else:
            Queue.Queue.__init__(self, maxthreads)
        self.__closed = False
        self._init_(maxthreads, can_grow)
        for i in range(0, maxthreads): RunThread(self)

    def __del__(self):
        if not self.__closed: self.close()
        
    def close(self):
        """ Close the ThreadPool object"""
        self.__closed = True
        for i in range(0, self.maxthreads):
            thr = self.get()
            thr.stop()

class ThreadPoolShared(ThreadPoolMixIn):
    """ A Fake ThreadPool that share Thread objects (RunThread)
    between ThreadPool's objects. Useful to save resources """
    def __init__(self, maxthreads, queue, can_grow = False):
        self._init_(maxthreads, can_grow)
        self.queue = queue
        self.get = self.queue.get
        self.put = self.queue.put
