#coding: iso-8859-1;
############################################################################
#                                                                          #
#    This file 'warn/__init__.py'                                          #
#    is part of 'SPyRO: Simple Python Remote Objects'                      #
#    Copyright (C) 2004-2005 by Eric Sadit Téllez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################


__all__ = ['debug','log','warn','stack','halt']
""" Debug and warning system (just error display). """

import traceback
import sys
import time
import os

class FileWriter:
    def __init__(self, name, outfile=None):
        self.name = name
        if outfile:
            self.file = outfile
        else:
            self.file = file(name,'a+b',0)

    def write(self, string):
        self.file.write(string)
    
class SyslogWriter:
    def __init__(self, name):
        self.name = name

    def write(self, string):
        syslog.openlog(self.name)
        syslog.syslog(syslog.LOG_ERR, string)
        syslog.closelog()


# In output_handlers is stored every special handler to manage the output
# the handlers must support write and _write methods, both of them must
# receive a string as argument.

output = {}
class Warn:
    LOG_SYSLOG = 1
    LOG_FILE = 2
    LOG_STDERR = 3
    output = output
    def __init__(self, name, outfile=None, filename=None, logtype=None):
        """ Creates a Warn object. Warn objects provide useful
        tools to show errors and warnings.
        |name| The name of the warn object
        |outfile| An output object in case that you have a default output.
        If None, a file will be created
        |logtype| If it must use syslog. If none it uses output to file.
        Possible values are: Warn.LOG_FILE, Warn.LOG_SYSLOG, Warn.LOG_STDERR
        Default: LOG_STDERR
        If LOG_FILE is specified the constructor needs any of the
        following keyword arguments:
        filename='filename-to-output' or outfile=fileobject

        NOTE: If output already has a handler, this handler is used overriding
        any of the opening instructions. In order to override, you need
        to modify the handler (using warn.output) directly
        """
        self.name = name
        if logtype is None: logtype = Warn.LOG_STDERR
        #### To prevent overwriting a previously defined handler
        if Warn.output.has_key(name): return
        if logtype == Warn.LOG_STDERR:
            Warn.output[name] = sys.stderr
        elif logtype == Warn.LOG_SYSLOG:
            Warn.output[name] = SyslogWriter(name)
        elif logtype == Warn.LOG_FILE:
            if not outfile:
                if not filename:
                    raise "You must specify a filename if you want a LOG_FILE Warn object"
                outfile = open(filename,'a+b',0)
            Warn.output[name] = outfile
        else:
            raise "Invalid type of Warn"
    
    def write(self, string):
        """ Output for loggin, append the current time, the name of the log, and the pid """
        self._write("%s %s: %s"%(time.ctime(time.time()), self.name, string))

    def _write(self, string):
        """ Direct write to the output """
        Warn.output[self.name].write(string)
        
    def debug(self, *args):
        """ Shows the args on the stderr """
        self.write("DEBUG: " + "".join(map(str,args)) + "\n")
        
    def log(self, *args):
        """ Shows the args on a log """
        self.write("LOG: " + "".join(map(str,args)) + "\n")
        
    def warn(self,*args):
        """ Shows |args| as warning, and print the last exception catched """
        self.write("WARNING: " + "".join(map(str,args)) + "\n")
        traceback.print_exc(file=self)
 
    def stack(self,*args):
        """ Shows |args| as warning, and print the stack of current caller """
        self.write("STACK: " + "".join(map(str,args)) + "\n")
        traceback.print_stack(file=self)

    def halt(self,*args):
        """ Displays |args|, print the last_exception and exits with code 127"""
        self.write("HALT: "  + "".join(map(str,args)) + "\n")
        traceback.print_exc(file=self)
        sys.exit(127)

_w = Warn("PID " + str(os.getpid()),outfile=sys.stderr)
debug = _w.debug
log = _w.log
warn = _w.warn
stack = _w.stack
halt = _w.halt
write = _w.write
_write = _w._write
